import numpy as np

class FDI_Annotator:
    def __init__(self,):
        self.inverse = {
            "Background": {"quadrant":0, "ID": "00", "type": "Background", "name": "Background"},
            "Lower Jawbone": {"quadrant": 0, "ID": "01", "type": "Lower Jawbone", "name": "Lower Jawbone"},
            "Upper Jawbone": {"quadrant": 0, "ID": "02", "type": "Upper Jawbone", "name": "Upper Jawbone"},
            "Left Inferior Alveolar Canal": {"quadrant": 0, "ID": "03", "type": "Left Inferior Alveolar Canal", "name": "Left Inferior Alveolar Canal"},
            "Right Inferior Alveolar Canal": {"quadrant": 0, "ID": "04", "type": "Right Inferior Alveolar Canal", "name": "Right Inferior Alveolar Canal"},
            "Left Maxillary Sinus": {"quadrant": 0, "ID": "05", "type": "Left Maxillary Sinus", "name": "Left Maxillary Sinus"},
            "Right Maxillary Sinus": {"quadrant": 0, "ID": "06", "type": "Right Maxillary Sinus", "name": "Right Maxillary Sinus"},
            "Pharynx": {"quadrant": 0, "ID": "07", "type": "Pharynx", "name": "Pharynx"},
            "Bridge": {"quadrant": 0, "ID": "08", "type": "Bridge", "name": "Bridge"},
            "Crown": {"quadrant": 0, "ID": "09", "type": "Crown", "name": "Crown"},
            "Implant": {"quadrant": 0, "ID": "10", "type": "Implant", "name": "Implant"},
            "Upper Right Central Incisor": {"quadrant": 1, "ID": "11", "type": "Incisor", "name": "Upper Right Central Incisor"},
            "Upper Right Lateral Incisor": {"quadrant": 1, "ID": "12", "type": "Incisor", "name": "Upper Right Lateral Incisor"},
            "Upper Right Canine": {"quadrant": 1, "ID": "13", "type": "Canine", "name": "Upper Right Canine"},
            "Upper Right First Premolar": {"quadrant": 1, "ID": "14", "type": "Premolar", "name": "Upper Right First Premolar"},
            "Upper Right Second Premolar": {"quadrant": 1, "ID": "15", "type": "Premolar", "name": "Upper Right Second Premolar"},
            "Upper Right First Molar": {"quadrant": 1, "ID": "16", "type": "Molar", "name": "Upper Right First Molar"},
            "Upper Right Second Molar": {"quadrant": 1, "ID": "17", "type": "Molar", "name": "Upper Right Second Molar"},
            "Upper Right Third Molar (Wisdom Tooth)": {"quadrant": 1, "ID": "18", "type": "Molar", "name": "Upper Right Third Molar (Wisdom Tooth)"},
            "Upper Left Central Incisor": {"quadrant": 2, "ID": "21", "type": "Incisor", "name": "Upper Left Central Incisor"},
            "Upper Left Lateral Incisor": {"quadrant": 2, "ID": "22", "type": "Incisor", "name": "Upper Left Lateral Incisor"},
            "Upper Left Canine": {"quadrant": 2, "ID": "23", "type": "Canine", "name": "Upper Left Canine"},
            "Upper Left First Premolar": {"quadrant": 2, "ID": "24", "type": "Premolar", "name": "Upper Left First Premolar"},
            "Upper Left Second Premolar": {"quadrant": 2, "ID": "25", "type": "Premolar", "name": "Upper Left Second Premolar"},
            "Upper Left First Molar": {"quadrant": 2, "ID": "26", "type": "Molar", "name": "Upper Left First Molar"},
            "Upper Left Second Molar": {"quadrant": 2, "ID": "27", "type": "Molar", "name": "Upper Left Second Molar"},
            "Upper Left Third Molar (Wisdom Tooth)": {"quadrant": 2, "ID": "28", "type": "Molar", "name": "Upper Left Third Molar (Wisdom Tooth)"},
            "Lower Left Central Incisor": {"quadrant": 3, "ID": "31", "type": "Incisor", "name": "Lower Left Central Incisor"},
            "Lower Left Lateral Incisor": {"quadrant": 3, "ID": "32", "type": "Incisor", "name": "Lower Left Lateral Incisor"},
            "Lower Left Canine": {"quadrant": 3, "ID": "33", "type": "Canine", "name": "Lower Left Canine"},
            "Lower Left First Premolar": {"quadrant": 3, "ID": "34", "type": "Premolar", "name": "Lower Left First Premolar"},
            "Lower Left Second Premolar": {"quadrant": 3, "ID": "35", "type": "Premolar", "name": "Lower Left Second Premolar"},
            "Lower Left First Molar": {"quadrant": 3, "ID": "36", "type": "Molar", "name": "Lower Left First Molar"},
            "Lower Left Second Molar": {"quadrant": 3, "ID": "37", "type": "Molar", "name": "Lower Left Second Molar"},
            "Lower Left Third Molar (Wisdom Tooth)": {"quadrant": 3, "ID": "38", "type": "Molar", "name": "Lower Left Third Molar (Wisdom Tooth)"},
            "Lower Right Central Incisor": {"quadrant": 4, "ID": "41", "type": "Incisor", "name": "Lower Right Central Incisor"},
            "Lower Right Lateral Incisor": {"quadrant": 4, "ID": "42", "type": "Incisor", "name": "Lower Right Lateral Incisor"},
            "Lower Right Canine": {"quadrant": 4, "ID": "43", "type": "Canine", "name": "Lower Right Canine"},
            "Lower Right First Premolar": {"quadrant": 4, "ID": "44", "type": "Premolar", "name": "Lower Right First Premolar"},
            "Lower Right Second Premolar": {"quadrant": 4, "ID": "45", "type": "Premolar", "name": "Lower Right Second Premolar"},
            "Lower Right First Molar": {"quadrant": 4, "ID": "46", "type": "Molar", "name": "Lower Right First Molar"},
            "Lower Right Second Molar": {"quadrant": 4, "ID": "47", "type": "Molar", "name": "Lower Right Second Molar"},
            "Lower Right Third Molar (Wisdom Tooth)": {"quadrant": 4, "ID": "48", "type": "Molar", "name": "Lower Right Third Molar (Wisdom Tooth)"}
        }


        self.fdi_notation = {
            "00": {"quadrant":0, "ID": "00", "type": "Background", "name": "Background"},
            "01": {"quadrant": 0, "ID": "01", "type": "Lower Jawbone", "name": "Lower Jawbone"},
            "02": {"quadrant": 0, "ID": "02", "type": "Upper Jawbone", "name": "Upper Jawbone"},
            "03": {"quadrant": 0, "ID": "03", "type": "Left Inferior Alveolar Canal", "name": "Left Inferior Alveolar Canal"},
            "04": {"quadrant": 0, "ID": "04", "type": "Right Inferior Alveolar Canal", "name": "Right Inferior Alveolar Canal"},
            "05": {"quadrant": 0, "ID": "05", "type": "Left Maxillary Sinus", "name": "Left Maxillary Sinus"},
            "06": {"quadrant": 0, "ID": "06", "type": "Right Maxillary Sinus", "name": "Right Maxillary Sinus"},
            "07": {"quadrant": 0, "ID": "07", "type": "Pharynx", "name": "Pharynx"},

            "11": {"quadrant": 1, "ID": "11", "type": "Incisor", "name": "Upper Right Central Incisor"},
            "12": {"quadrant": 1, "ID": "12", "type": "Incisor", "name": "Upper Right Lateral Incisor"},
            "13": {"quadrant": 1, "ID": "13", "type": "Canine", "name": "Upper Right Canine"},
            "14": {"quadrant": 1, "ID": "14", "type": "Premolar", "name": "Upper Right First Premolar"},
            "15": {"quadrant": 1, "ID": "15", "type": "Premolar", "name": "Upper Right Second Premolar"},
            "16": {"quadrant": 1, "ID": "16", "type": "Molar", "name": "Upper Right First Molar"},
            "17": {"quadrant": 1, "ID": "17", "type": "Molar", "name": "Upper Right Second Molar"},
            "18": {"quadrant": 1, "ID": "18", "type": "Molar", "name": "Upper Right Third Molar (Wisdom Tooth)"},

            "21": {"quadrant": 2, "ID": "21", "type": "Incisor", "name": "Upper Left Central Incisor"},
            "22": {"quadrant": 2, "ID": "22", "type": "Incisor", "name": "Upper Left Lateral Incisor"},
            "23": {"quadrant": 2, "ID": "23", "type": "Canine", "name": "Upper Left Canine"},
            "24": {"quadrant": 2, "ID": "24", "type": "Premolar", "name": "Upper Left First Premolar"},
            "25": {"quadrant": 2, "ID": "25", "type": "Premolar", "name": "Upper Left Second Premolar"},
            "26": {"quadrant": 2, "ID": "26", "type": "Molar", "name": "Upper Left First Molar"},
            "27": {"quadrant": 2, "ID": "27", "type": "Molar", "name": "Upper Left Second Molar"},
            "28": {"quadrant": 2, "ID": "28", "type": "Molar", "name": "Upper Left Third Molar (Wisdom Tooth)"},

            "31": {"quadrant": 3, "ID": "31", "type": "Incisor", "name": "Lower Left Central Incisor"},
            "32": {"quadrant": 3, "ID": "32", "type": "Incisor", "name": "Lower Left Lateral Incisor"},
            "33": {"quadrant": 3, "ID": "33", "type": "Canine", "name": "Lower Left Canine"},
            "34": {"quadrant": 3, "ID": "34", "type": "Premolar", "name": "Lower Left First Premolar"},
            "35": {"quadrant": 3, "ID": "35", "type": "Premolar", "name": "Lower Left Second Premolar"},
            "36": {"quadrant": 3, "ID": "36", "type": "Molar", "name": "Lower Left First Molar"},
            "37": {"quadrant": 3, "ID": "37", "type": "Molar", "name": "Lower Left Second Molar"},
            "38": {"quadrant": 3, "ID": "38", "type": "Molar", "name": "Lower Left Third Molar (Wisdom Tooth)"},

            "41": {"quadrant": 4, "ID": "41", "type": "Incisor", "name": "Lower Right Central Incisor"},
            "42": {"quadrant": 4, "ID": "42", "type": "Incisor", "name": "Lower Right Lateral Incisor"},
            "43": {"quadrant": 4, "ID": "43", "type": "Canine", "name": "Lower Right Canine"},
            "44": {"quadrant": 4, "ID": "44", "type": "Premolar", "name": "Lower Right First Premolar"},
            "45": {"quadrant": 4, "ID": "45", "type": "Premolar", "name": "Lower Right Second Premolar"},
            "46": {"quadrant": 4, "ID": "46", "type": "Molar", "name": "Lower Right First Molar"},
            "47": {"quadrant": 4, "ID": "47", "type": "Molar", "name": "Lower Right Second Molar"},
            "48": {"quadrant": 4, "ID": "48", "type": "Molar", "name": "Lower Right Third Molar (Wisdom Tooth)"}
        }


if __name__ == "__main__":
    annot = FDI_Annotator()
    for k, name in annot.fdi_notation.items():
        print(int(k), name['name'])



import os
import numpy as np


def napari_get_reader(path):
    print(path)
    if isinstance(path, str) and os.path.isdir(path):
        return reader_function
    return None


def reader_function(path):
    annotation_path = os.path.join(path, 'annotation.npy')
    volume_path = os.path.join(path, 'volume.npy')

    annotation = np.load(annotation_path)
    volume = np.load(volume_path)
    metadata = {'parent_folder': path}

    return [
        (
            volume, 
            {
                'name': 'volume',
                'metadata': metadata,
            },
            'image'
        ), (
            annotation,
            {
                'name': 'annotation',
                'metadata': metadata
            },
            'labels'
        )
    ]



import numpy as np

from napari_toothfairy_annotator import napari_get_reader


# tmp_path is a pytest fixture
def test_reader(tmp_path):
    """An example of how you might test your plugin."""

    # write some fake data using your supported file format
    my_test_file = str(tmp_path / "myfile.npy")
    original_data = np.random.rand(20, 20)
    np.save(my_test_file, original_data)

    # try to read it back in
    reader = napari_get_reader(my_test_file)
    assert callable(reader)

    # make sure we're delivering the right format
    layer_data_list = reader(my_test_file)
    assert isinstance(layer_data_list, list) and len(layer_data_list) > 0
    layer_data_tuple = layer_data_list[0]
    assert isinstance(layer_data_tuple, tuple) and len(layer_data_tuple) > 0

    # make sure it's the same as it started
    np.testing.assert_allclose(original_data, layer_data_tuple[0])


def test_get_reader_pass():
    reader = napari_get_reader("fake.file")
    assert reader is None



import numpy as np

from napari_toothfairy_annotator._widget import (
    ExampleQWidget,
    ImageThreshold,
    threshold_autogenerate_widget,
    threshold_magic_widget,
)


def test_threshold_autogenerate_widget():
    # because our "widget" is a pure function, we can call it and
    # test it independently of napari
    im_data = np.random.random((100, 100))
    thresholded = threshold_autogenerate_widget(im_data, 0.5)
    assert thresholded.shape == im_data.shape
    # etc.


# make_napari_viewer is a pytest fixture that returns a napari viewer object
# you don't need to import it, as long as napari is installed
# in your testing environment
def test_threshold_magic_widget(make_napari_viewer):
    viewer = make_napari_viewer()
    layer = viewer.add_image(np.random.random((100, 100)))

    # our widget will be a MagicFactory or FunctionGui instance
    my_widget = threshold_magic_widget()

    # if we "call" this object, it'll execute our function
    thresholded = my_widget(viewer.layers[0], 0.5)
    assert thresholded.shape == layer.data.shape
    # etc.


def test_image_threshold_widget(make_napari_viewer):
    viewer = make_napari_viewer()
    layer = viewer.add_image(np.random.random((100, 100)))
    my_widget = ImageThreshold(viewer)

    # because we saved our widgets as attributes of the container
    # we can set their values without having to "interact" with the viewer
    my_widget._image_layer_combo.value = layer
    my_widget._threshold_slider.value = 0.5

    # this allows us to run our functions directly and ensure
    # correct results
    my_widget._threshold_im()
    assert len(viewer.layers) == 2


# capsys is a pytest fixture that captures stdout and stderr output streams
def test_example_q_widget(make_napari_viewer, capsys):
    # make viewer and add an image layer using our fixture
    viewer = make_napari_viewer()
    viewer.add_image(np.random.random((100, 100)))

    # create our widget, passing in the viewer
    my_widget = ExampleQWidget(viewer)

    # call our widget method
    my_widget._on_click()

    # read captured output and check that it's as we expected
    captured = capsys.readouterr()
    assert captured.out == "napari has 1 layers\n"



__version__ = "0.0.6"

from ._reader import napari_get_reader
from ._widget import WidgetAnnotator
from ._writer import write_multiple, write_single_image
from ._widget import FolderBrowser

__all__ = (
    "napari_get_reader",
    "write_single_image",
    "write_multiple",
    "WidgetAnnotator",
    "FolderBrowser"
)



"""
This module is an example of a barebones writer plugin for napari.

It implements the Writer specification.
see: https://napari.org/stable/plugins/guides.html?#writers

Replace code below according to your needs.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Sequence, Tuple, Union

if TYPE_CHECKING:
    DataType = Union[Any, Sequence[Any]]
    FullLayerData = Tuple[DataType, dict, str]


def write_single_image(path: str, data: Any, meta: dict) -> List[str]:
    """Writes a single image layer.
    
    Parameters
    ----------
    path : str
        A string path indicating where to save the image file.
    data : The layer data
        The `.data` attribute from the napari layer.
    meta : dict
        A dictionary containing all other attributes from the napari layer
        (excluding the `.data` layer attribute).

    Returns
    -------
    [path] : A list containing the string path to the saved file.
    """

    # implement your writer logic here ...

    # return path to any file(s) that were successfully written
    return [path]


def write_multiple(path: str, data: List[FullLayerData]) -> List[str]:
    """Writes multiple layers of different types.
    
    Parameters
    ----------
    path : str
        A string path indicating where to save the data file(s).
    data : A list of layer tuples.
        Tuples contain three elements: (data, meta, layer_type)
        `data` is the layer data
        `meta` is a dictionary containing all other metadata attributes
        from the napari layer (excluding the `.data` layer attribute).
        `layer_type` is a string, eg: "image", "labels", "surface", etc.

    Returns
    -------
    [path] : A list containing (potentially multiple) string paths to the saved file(s).
    """

    # implement your writer logic here ...

    # return path to any file(s) that were successfully written
    return [path]



import os
import json
import numpy as np
from pathlib import Path
# from napari_plugin_engine import napari_hook_implementation

from qtpy.QtWidgets import (
    QAbstractItemView,
    QComboBox,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListWidget,
    QListWidgetItem,
    QMenu,
    QPushButton,
    QTreeView,
    QVBoxLayout,
    QWidget,
    QFileSystemModel
)
from qtpy.QtCore import  (
    QDir,
    QModelIndex,
    QPoint,
    QRegExp,
    QSortFilterProxyModel,
    Qt,
)
from qtpy.QtGui import (
    QColor,
    QPainter,
    QIcon,
)

from napari.viewer import Viewer
from magicgui.widgets import FileEdit
from magicgui.types import FileDialogMode

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from qtpy.QtCore import QModelIndex

from napari_toothfairy_annotator.FDI_Annotator import FDI_Annotator

class ColorWidgetItem(QListWidgetItem):
    def __init__(self, text, color):
        super().__init__(text)
        self.color = color

    def paint(self, painter):
        painter.fillRect(self.rect().adjusted(0, 0, -50, 0), self.color)
        super().paint(painter)


class WidgetAnnotator(QWidget):
    def __init__(self, napari_viewer):
        super().__init__()
        self.viewer = napari_viewer
        self.fdi_annotator = FDI_Annotator()
        self.associations = {0: "00"}
        self._available_ids = None

        self.load_associated_volume()

        self.viewer.layers.move_multiple([0,2,1])

        layout = QVBoxLayout()

        self.list1 = QListWidget()
        # self.list1.setSortingEnabled(True)
        layout.addWidget(self.list1)

        self.list2 = QListWidget()
        self.list2.setSortingEnabled(True)
        layout.addWidget(self.list2)
        self.list2.setSelectionMode(
            QAbstractItemView.ExtendedSelection
        )

        self.associate_button = QPushButton("Associate IDs")
        self.associate_button.clicked.connect(self.associate_ids)
        layout.addWidget(self.associate_button)

        self.reload_button = QPushButton("Reload")
        self.reload_button.clicked.connect(self.reload)
        layout.addWidget(self.reload_button)

        self.setLayout(layout)
        self.load_associations()
        self.update_lists()


    def reload(self,):
        source = self.get_source()
        annotation_npy_path = os.path.join(source, 'annotation.npy')

        if not os.path.isfile(annotation_npy_path):
            print('path not found on reload')
            return

        self.load_associations()
        self.viewer.layers.remove('annotation')
        annotation = np.load(annotation_npy_path)

        keys = self.associations.keys()
        print(keys)
        for key in keys:
            mask = annotation == int(key)
            annotation[mask] = 0

        self.viewer.add_labels(annotation, name='annotation', visible=True)


    def load_associated_volume(self,):
        source = self.get_source()
        associated_volume_path = os.path.join(source, 'associated.npy')
        if os.path.isfile(associated_volume_path):
            self.associated_volume = np.load(associated_volume_path)
        else:
            self.associated_volume = np.zeros_like(self.viewer.layers['annotation'].data)
        self.viewer.add_labels(self.associated_volume, name='associated', visible=False)

    def associate_ids(self):
        selected_items_list1 = self.list1.selectedItems()
        selected_items_list2 = self.list2.selectedItems()

        for item1 in selected_items_list1:
            for item2 in selected_items_list2:
                fdi_id = self.fdi_annotator.inverse[item1.text()]['ID']
                self.associations[int(item2.text())] = fdi_id
                mask = self.viewer.layers['annotation'].data == int(item2.text())
                self.viewer.layers['annotation'].data[mask] = 0
                self.viewer.layers['associated'].data[mask] = int(fdi_id)
        self.viewer.layers['annotation'].refresh()
        self.viewer.layers['associated'].refresh()
        self.update_lists()
        self.save()

    def get_source(self,):
        source = self.viewer.layers['annotation'].source.path
        if source is None:
            source = self.viewer.layers['volume'].source.path
        if source is None:
            source = self.viewer.layers['annotation'].metadata['parent_folder']
        if source is None:
            source = self.viewer.layers['volume'].metadata['parent_folder']

        assert source is not None, "I tried so hard and got so far but in the end it doesn't even matter"

        return source


    def save(self,):
        data = self.viewer.layers['associated'].data
        source = self.get_source()

        save_path = os.path.join(source, 'associated.npy')
        np.save(save_path, data)
        association_file = os.path.join(source, 'associations.json')
        with open(association_file, 'w') as f:
            json.dump(self.associations, f)


    def load_associations(self,):
        source = self.get_source()
        association_file = os.path.join(source, 'associations.json')

        if not os.path.isfile(association_file):
            return

        with open(association_file) as f:
            print('Load associations.json')
            str_id_associations = json.load(f)
            self.associations = {}
            for k, v in str_id_associations.items():
                self.associations[int(k)] = v

        for left_id, right_id in self.associations.items():
            mask = self.viewer.layers['annotation'].data == int(left_id)
            self.viewer.layers['annotation'].data[mask] = 0
        self.viewer.layers['annotation'].refresh()


    def update_lists(self):
        self.list1.clear()
        self.list2.clear()

        for id_data in self.get_fdi_ids():
            item = ColorWidgetItem(id_data['name'], QColor("white"))
            self.list1.addItem(item)

        already_annotated = set(self.associations.keys())
        print(f'already_annotated: {already_annotated}')

        for id_data in self.get_available_ids():
            s = f'{id_data}'
            if int(id_data) in already_annotated:
                assoc_id = self.associations[int(id_data)]
                s += f' > {self.fdi_annotator.fdi_notation[assoc_id]["name"]}'
            item = ColorWidgetItem(s, QColor("red"))
            self.list2.addItem(item)

    def get_available_ids(self,):
        if self._available_ids is None:
            data = self.viewer.layers['annotation'].data
            self._available_ids = np.unique(data).tolist()
            self._available_ids += self.associations.keys()
            self._available_ids = list(set(self._available_ids))
        return self._available_ids

    def get_fdi_ids(self,):
        return self.fdi_annotator.fdi_notation.values()

    def get_available_layers(self,):
        return [layer.name for layer in self.viewer.layers]


class DirectoryFriendlyFilterProxyModel(QSortFilterProxyModel):
    def filterAcceptsRow(self, source_row: int, source_parent: QModelIndex) -> bool:
        """Accepts directories and files that pass the base class's filter
        
        Note: This custom proxy ensures that we can search for filenames and keep the directories
        in the tree view
        """
        model = self.sourceModel()
        index = model.index(source_row, 0, source_parent)

        if model.isDir(index):
            return True

        return super().filterAcceptsRow(source_row, source_parent)


class FolderBrowser(QWidget):
    """Main Widget for the Folder Browser Dock Widget
    
    The napari viewer is passed in as an argument to the constructor
    """
    viewer: Viewer
    folder_chooser: FileEdit
    file_system_model: QFileSystemModel
    proxy_model: DirectoryFriendlyFilterProxyModel
    search_field: QLineEdit
    tree_view: QTreeView
    annotator_widget: WidgetAnnotator

    def __init__(self, napari_viewer):
        super().__init__()
        self.viewer = napari_viewer

        self.setLayout(QVBoxLayout())

        current_directory: Path = Path(QFileDialog.getExistingDirectory(self, "Select Directory"))
        self.layout().addWidget(QLabel("Directory"))
        self.folder_chooser = FileEdit(
            mode=FileDialogMode.EXISTING_DIRECTORY,
            value=current_directory,
        )
        self.layout().addWidget(self.folder_chooser.native)

        def directory_changed(*_) -> None:
            current_directory = Path(self.folder_chooser.value)
            self.tree_view.setRootIndex(
                self.proxy_model.mapFromSource(
                    self.file_system_model.index(current_directory.as_posix())
                )
            )

        self.folder_chooser.line_edit.changed.connect(directory_changed)

        # --------------------------------------------
        # File system abstraction with proxy for search filtering
        self.file_system_model = QFileSystemModel()
        self.file_system_model.setRootPath(QDir.rootPath())
        self.proxy_model = DirectoryFriendlyFilterProxyModel()
        self.proxy_model.setSourceModel(self.file_system_model)

        # Create search box and connect to proxy model
        self.layout().addWidget(QLabel("File filter"))
        self.search_field = QLineEdit()
        # Note: We should agree on the best regex interaction to provide here
        def update_filter(text: str) -> None:
            self.proxy_model.setFilterRegExp(QRegExp(text, Qt.CaseInsensitive))
        self.search_field.textChanged.connect(update_filter)
        search_widget = QWidget()
        search_widget.setLayout(QHBoxLayout())
        search_widget.layout().addWidget(QLabel("Search:"))
        search_widget.layout().addWidget(self.search_field)
        self.layout().addWidget(search_widget)

        self.tree_view = QTreeView()
        self.tree_view.setModel(self.proxy_model)
        self.tree_view.setRootIndex(
            self.proxy_model.mapFromSource(
                self.file_system_model.index(current_directory.as_posix())
            )
        )
    
        self.tree_view.doubleClicked.connect(self.__tree_double_click)

        self.tree_view.setHeaderHidden(True)
        self.tree_view.hideColumn(1)
        self.tree_view.hideColumn(2)
        self.tree_view.hideColumn(3)
        
        self.layout().addWidget(self.tree_view)

        self.annotator_widget = None


    def __tree_double_click(self, index: QModelIndex) -> None:
        """Action on double click in the tree model
        
        Opens the selected file or in the folder
        
        Args:
            index: Index of the selected item in the tree view
        """

        source_index: QModelIndex = self.proxy_model.mapToSource(index)
        file_path: str = self.file_system_model.filePath(source_index)

        if self.file_system_model.isDir(source_index):
            if self.annotator_widget is not None:
                self.layout().removeWidget(self.annotator_widget)
                self.annotator_widget = None

            layers_to_remove = self.viewer.layers.copy()
            for layer in layers_to_remove:
                self.viewer.layers.remove(layer)

            print(f'Layers: {len(self.viewer.layers)}')

            self.viewer.open(file_path, plugin='napari-toothfairy-annotator')
            self.annotator_widget = WidgetAnnotator(self.viewer)
            self.layout().addWidget(self.annotator_widget)

    def __show_context_menu(self, position: QPoint) -> None:
        """Show a context menu when right-clicking in the tree view"""
        menu = QMenu()
        open_multiple_action = menu.addAction("Open multiple files")
        open_multiple_action.triggered.connect(
            lambda: self.__open_multi_selection(is_stack=False)
        )
        open_as_stack_action = menu.addAction("Open as stack")
        open_as_stack_action.triggered.connect(
            lambda: self.__open_multi_selection(is_stack=True)
        )

        menu.exec_(self.tree_view.viewport().mapToGlobal(position))

    def __open_multi_selection(self, is_stack: bool) -> None:
        """Open multiple files in the viewer
        
        The files are selected in the tree view
        
        Args:
            is_stack: If True, the files are opened as a stack
        """
        indices: list[QModelIndex] = self.tree_view.selectionModel().selectedIndexes()
        fs_paths: list[str] = [
            self.file_system_model.filePath(self.proxy_model.mapToSource(index)) for index in indices
            if not self.file_system_model.isDir(self.proxy_model.mapToSource(index)) and index.column() == 0
        ]

        # Nothing to do when there is no file selected
        if len(fs_paths) == 0:
            return

        self.viewer.open(fs_paths, stack=is_stack)


# @napari_hook_implementation
# def napari_experimental_provide_dock_widget():
#     return [FolderBrowser]
